//! Definitions and decoding of ATEM protocol fields

/// An uninterpreted ATEM protocol field - a 4-ascii character type plus variable length data
pub struct RawField<'a> {
    pub r#type: &'a str,
    pub data: &'a [u8],
}

#[derive(Debug)]
pub struct _Ver {
    pub major: u16,
    pub minor: u16,
}

impl<'a> Field<'a> for _Ver {
    const TYPE: &'static str = "_ver";

    fn decode(data: &'a [u8]) -> Result<Self, FieldParsingError> {
        let data = checked_len::<4>(data)?;
        Ok(Self {
            major: u16::from_be_bytes(data[0..=1].try_into().unwrap()),
            minor: u16::from_be_bytes(data[2..=3].try_into().unwrap()),
        })
    }
}

#[derive(Debug)]
pub struct PrvI {
    pub m_e_index: u8,
    pub source_index: u16,
    pub pvw_in_pgm: bool,
}

impl<'a> Field<'a> for PrvI {
    const TYPE: &'static str = "PrvI";
    fn decode(data: &'a [u8]) -> Result<Self, FieldParsingError> {
        let data = checked_len::<8>(data)?;
        Ok(Self {
            m_e_index: data[0],
            source_index: u16::from_be_bytes(data[2..=3].try_into().unwrap()),
            pvw_in_pgm: data[4] != 0,
        })
    }
}

#[derive(Debug)]
pub struct PrgI {
    pub m_e_index: u8,
    pub source_index: u16,
}

impl<'a> Field<'a> for PrgI {
    const TYPE: &'static str = "PrgI";
    fn decode(data: &'a [u8]) -> Result<Self, FieldParsingError> {
        let data = checked_len::<4>(data)?;
        Ok(Self {
            m_e_index: data[0],
            source_index: u16::from_be_bytes(data[2..=3].try_into().unwrap()),
        })
    }
}

pub trait Field<'a>: Sized {
    const TYPE: &'static str;
    fn decode(data: &'a [u8]) -> Result<Self, FieldParsingError>;
    fn try_from_raw(raw: RawField<'a>) -> Result<Self, FieldParsingError> {
        if Self::TYPE != raw.r#type {
            Err(FieldParsingError::MismatchedFieldType)
        } else {
            Self::decode(&raw.data)
        }
    }
}

#[derive(Debug)]
pub enum FieldParsingError {
    UnexpectedLength { expected: usize, got: usize },
    MismatchedFieldType,
}

fn checked_len<'a, const LEN: usize>(data: &'a [u8]) -> Result<&'a [u8; LEN], FieldParsingError> {
    data.try_into()
        .map_err(|_| FieldParsingError::UnexpectedLength {
            expected: LEN,
            got: data.len(),
        })
}
