use std::{collections::HashMap, fmt::Debug, process::Command, sync::Arc};

use crate::{enums::ProtocolVersion, state::AtemState};

pub trait DeserializedCommand: Send + Sync + Debug {
    fn raw_name(&self) -> &'static str;
    fn apply_to_state(&self, state: &mut AtemState);
}

pub trait CommandDeserializer: Send + Sync {
    fn deserialize(&self, buffer: &[u8], version: &ProtocolVersion)
        -> Arc<dyn DeserializedCommand>;
}

pub trait SerializableCommand: Send + Sync {
    fn payload(&self, version: &ProtocolVersion) -> Vec<u8>;
}

impl<C: SerializableCommand + ?Sized> SerializableCommand for Box<C> {
    fn payload(&self, version: &ProtocolVersion) -> Vec<u8> {
        (**self).payload(version)
    }
}

impl<C: SerializableCommand + ?Sized> SerializableCommand for &'_ Box<C> {
    fn payload(&self, version: &ProtocolVersion) -> Vec<u8> {
        (**self).payload(version)
    }
}

pub trait BasicWritableCommand: SerializableCommand + Send + Sync {
    fn get_raw_name(&self) -> &'static str;
    fn get_minimum_version(&self) -> ProtocolVersion;
}

impl<C: BasicWritableCommand + ?Sized> BasicWritableCommand for Box<C> {
    fn get_raw_name(&self) -> &'static str {
        (**self).get_raw_name()
    }

    fn get_minimum_version(&self) -> ProtocolVersion {
        (**self).get_minimum_version()
    }
}

impl<C: BasicWritableCommand + ?Sized> BasicWritableCommand for &'_ Box<C> {
    fn get_raw_name(&self) -> &'static str {
        (**self).get_raw_name()
    }

    fn get_minimum_version(&self) -> ProtocolVersion {
        (**self).get_minimum_version()
    }
}

pub trait WritableCommand: BasicWritableCommand {
    fn get_mask_flag(&self) -> HashMap<String, f64>;
    fn get_flag(&self) -> f64;
    fn set_flag(&mut self, flag: f64);
}
