use std::sync::Arc;

use crate::{
    commands::command_base::{CommandDeserializer, DeserializedCommand},
    enums::ProtocolVersion,
    state::info::{AtemCapabilites, MultiviewerInfo},
};

pub const DESERIALIZE_TOPOLOGY_RAW_NAME: &str = "_top";

#[derive(Debug)]
pub struct Topology {
    mix_effects: u8,
    sources: u8,
    auxilliaries: u8,
    mix_minus_outputs: u8,
    media_players: u8,
    multiviewers: Option<u8>,
    serial_ports: u8,
    max_hyperdecks: u8,
    dves: u8,
    stingers: u8,
    super_sources: u8,
    talkback_channels: u8,
    downstream_keyers: u8,
    camera_control: bool,
    advanced_chroma_keyers: bool,
    only_configurable_outputs: bool,
}

impl DeserializedCommand for Topology {
    fn raw_name(&self) -> &'static str {
        DESERIALIZE_TOPOLOGY_RAW_NAME
    }

    fn apply_to_state(&self, state: &mut crate::state::AtemState) {
        state.info.capabilities = Some(AtemCapabilites::new(
            self.mix_effects,
            self.sources,
            self.auxilliaries,
            self.mix_minus_outputs,
            self.media_players,
            self.serial_ports,
            self.max_hyperdecks,
            self.dves,
            self.stingers,
            self.super_sources,
            self.talkback_channels,
            self.downstream_keyers,
            self.camera_control,
            self.advanced_chroma_keyers,
            self.only_configurable_outputs,
        ));

        let window_count = if let Some(mv) = &state.info.multiviewer {
            *mv.window_count()
        } else {
            10
        };

        state.info.multiviewer = Some(MultiviewerInfo::new(self.multiviewers, window_count));
    }
}

#[derive(Default)]
pub struct TopologyDeserializer {}

impl CommandDeserializer for TopologyDeserializer {
    fn deserialize(
        &self,
        buffer: &[u8],
        version: &ProtocolVersion,
    ) -> Arc<dyn DeserializedCommand> {
        let v230offset = if *version > ProtocolVersion::V8_0_1 {
            1
        } else {
            0
        };

        let multiviewers = if v230offset > 0 {
            Some(buffer[6])
        } else {
            None
        };

        let advanced_chroma_keyers = if buffer.len() > 20 {
            buffer[21 + v230offset] == 1
        } else {
            false
        };

        let only_configurable_outputs = if buffer.len() > 20 {
            buffer[22 + v230offset] == 1
        } else {
            false
        };

        Arc::new(Topology {
            mix_effects: buffer[0],
            sources: buffer[1],
            downstream_keyers: buffer[2],
            auxilliaries: buffer[3],
            mix_minus_outputs: buffer[4],
            media_players: buffer[5],
            multiviewers,
            serial_ports: buffer[6 + v230offset],
            max_hyperdecks: buffer[7 + v230offset],
            dves: buffer[8 + v230offset],
            stingers: buffer[9 + v230offset],
            super_sources: buffer[10 + v230offset],
            talkback_channels: buffer[12 + v230offset],
            camera_control: buffer[17 + v230offset] == 1,
            advanced_chroma_keyers,
            only_configurable_outputs,
        })
    }
}
