use std::sync::Arc;

use crate::{
    commands::command_base::{
        BasicWritableCommand, CommandDeserializer, DeserializedCommand, SerializableCommand,
    },
    enums::ProtocolVersion,
    state::util::get_mix_effect,
};

pub const DESERIALIZE_PROGRAM_INPUT_RAW_NAME: &str = "PrgI";

#[derive(Debug, new)]
pub struct ProgramInput {
    pub mix_effect: u8,
    pub source: u16,
}

impl SerializableCommand for ProgramInput {
    fn payload(&self, _version: &crate::enums::ProtocolVersion) -> Vec<u8> {
        let mut buf = vec![0; 4];
        buf[..1].copy_from_slice(&self.mix_effect.to_be_bytes());
        buf[2..].copy_from_slice(&self.source.to_be_bytes());

        buf
    }
}

impl BasicWritableCommand for ProgramInput {
    fn get_raw_name(&self) -> &'static str {
        "CPgI"
    }

    fn get_minimum_version(&self) -> crate::enums::ProtocolVersion {
        crate::enums::ProtocolVersion::Unknown
    }
}

impl DeserializedCommand for ProgramInput {
    fn raw_name(&self) -> &'static str {
        DESERIALIZE_PROGRAM_INPUT_RAW_NAME
    }

    fn apply_to_state(&self, state: &mut crate::state::AtemState) {
        let Some(capabilities) = state.info.capabilities() else {
            todo!("Return error");
        };

        if self.mix_effect > *capabilities.mix_effects() {
            todo!("Return error");
        }

        let mix_effect = get_mix_effect(state, self.mix_effect as usize);
        mix_effect.program_input = self.source;
    }
}

#[derive(Default)]
pub struct ProgramInputDeserializer {}

impl CommandDeserializer for ProgramInputDeserializer {
    fn deserialize(
        &self,
        buffer: &[u8],
        version: &ProtocolVersion,
    ) -> Arc<dyn DeserializedCommand> {
        let mix_effect = buffer[0];
        let source = u16::from_be_bytes([buffer[2], buffer[3]]);

        Arc::new(ProgramInput { mix_effect, source })
    }
}
