use std::{collections::VecDeque, sync::Arc};

use crate::{
    atem_lib::atem_packet::{AtemPacket, RawFields},
    commands::device_profile::version::{deserialize_version, DESERIALIZE_VERSION_RAW_NAME},
    enums::ProtocolVersion,
};

use super::{
    command_base::{CommandDeserializer, DeserializedCommand},
    device_profile::{
        audio_mixer_config::{AudioMixerConfigDeserializer, DESERIALIZE_AUDIO_MIXER_CONFIG_NAME},
        media_pool_config::{MediaPoolConfigDeserializer, DESERIALIZE_MEDIA_POOL_CONFIG_NAME},
        mix_effect_block_config::{
            MixEffectBlockConfigDeserializer, DESERIALIZE_MIX_EFFECT_BLOCK_CONFIG_NAME,
        },
        multiviewer_config::{MultiviewerConfigDeserializer, DESERIALIZE_MULTIVIEWER_NAME},
        product_identifier::{
            ProductIdentifierDeserializer, DESERIALIZE_PRODUCT_IDENTIFIER_RAW_NAME,
        },
        topology::{TopologyDeserializer, DESERIALIZE_TOPOLOGY_RAW_NAME},
    },
    init_complete::{InitCompleteDeserializer, DESERIALIZE_INIT_COMPLETE_RAW_NAME},
    mix_effects::program_input::{ProgramInputDeserializer, DESERIALIZE_PROGRAM_INPUT_RAW_NAME},
    tally_by_source::{TallyBySourceDeserializer, DESERIALIZE_TALLY_BY_SOURCE_RAW_NAME},
    time::{TimeDeserializer, DESERIALIZE_TIME_RAW_NAME},
};

pub fn deserialize_commands<T: AsRef<[u8]>>(
    payload: T,
    version: &mut ProtocolVersion,
) -> VecDeque<Arc<dyn DeserializedCommand>> {
    let mut parsed_commands: VecDeque<Arc<dyn DeserializedCommand>> = VecDeque::new();

    for field in RawFields::new(payload.as_ref()) {
        let name: &str = field.r#type.try_into().unwrap();
        log::debug!("Received command {} with length {}", name, field.data.len(),);

        if field.r#type == DESERIALIZE_VERSION_RAW_NAME {
            let version_command = deserialize_version(field.data);
            *version = version_command.version.clone();
            log::info!("Switched to protocol version {}", version);
            parsed_commands.push_back(Arc::new(version_command));
        } else if let Some(deserializer) = command_deserializer_from_string(name) {
            let deserialized_command = deserializer.deserialize(field.data, version);
            log::debug!("Received {:?}", deserialized_command);
            parsed_commands.push_back(deserialized_command);
        } else {
            log::warn!("Received command {name} for which there is no deserializer.");
        }
    }

    parsed_commands
}

fn command_deserializer_from_string(command_str: &str) -> Option<Box<dyn CommandDeserializer>> {
    match command_str {
        DESERIALIZE_INIT_COMPLETE_RAW_NAME => Some(Box::<InitCompleteDeserializer>::default()),
        DESERIALIZE_PROGRAM_INPUT_RAW_NAME => Some(Box::<ProgramInputDeserializer>::default()),
        DESERIALIZE_TALLY_BY_SOURCE_RAW_NAME => Some(Box::<TallyBySourceDeserializer>::default()),
        DESERIALIZE_TIME_RAW_NAME => Some(Box::<TimeDeserializer>::default()),
        DESERIALIZE_TOPOLOGY_RAW_NAME => Some(Box::<TopologyDeserializer>::default()),
        DESERIALIZE_MIX_EFFECT_BLOCK_CONFIG_NAME => {
            Some(Box::<MixEffectBlockConfigDeserializer>::default())
        }
        DESERIALIZE_PRODUCT_IDENTIFIER_RAW_NAME => {
            Some(Box::<ProductIdentifierDeserializer>::default())
        }
        DESERIALIZE_MEDIA_POOL_CONFIG_NAME => Some(Box::<MediaPoolConfigDeserializer>::default()),
        DESERIALIZE_MULTIVIEWER_NAME => Some(Box::<MultiviewerConfigDeserializer>::default()),
        DESERIALIZE_AUDIO_MIXER_CONFIG_NAME => Some(Box::<AudioMixerConfigDeserializer>::default()),
        _ => None,
    }
}
